/**
    biobambam
    Copyright (C) 2019 German Tischler

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/
#include <biobambam2/DepthInterval.hpp>

std::ostream & operator<<(std::ostream & out, biobambam2::DepthInterval const & D)
{
	out << D.refid << "\t" << D.from << "\t" << D.to << "\t" << D.depth;
	return out;
}

std::string biobambam2::DepthInterval::print(libmaus2::bambam::BamHeader const & header) const
{
	std::ostringstream out;
	out << refid << "\t" << header.getRefIDName(refid) << "\t" << from << "\t" << to << "\t" << depth;
	return out.str();
}

